# Copyright 2022 Baler Contributors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil

from baler.modules import helper


def test_create_new_project():
    workspace_name = "test_workspace"
    project_name = "test_project"
    base_path = os.path.join("test_workspaces")
    workspace_path = os.path.join(base_path, workspace_name)
    project_path = os.path.join(workspace_path, project_name)

    # Ensure the project doesn't exist yet
    if os.path.exists(workspace_path):
        shutil.rmtree(workspace_path)

    helper.create_new_project(workspace_name, project_name, base_path=base_path)

    # Verify that the project was created successfully
    assert os.path.exists(workspace_path)
    assert os.path.exists(project_path)

    required_directories = [
        os.path.join(workspace_path, "data"),
        os.path.join(project_path, "config"),
        os.path.join(project_path, "output", "compressed_output"),
        os.path.join(project_path, "output", "decompressed_output"),
        os.path.join(project_path, "output", "plotting"),
        os.path.join(project_path, "output", "training"),
    ]
    for directory in required_directories:
        print(directory)
        assert os.path.exists(directory)

    # Clean up after the test
    shutil.rmtree(base_path)



import os
import time
from math import ceil

import numpy as np

import baler_compressor.helper as helper
import gzip
import sys


def run(input_data_path, model_output_path, normalization_features, config):
    """Main function calling the compression functions, ran when --mode=compress is selected.
    The main function being called here is: `helper.compress`

        If `config.extra_compression` is selected, the compressed file is further compressed via zip
        Else, the function returns a compressed file of `.npz`, only compressed by Baler.

    Args:
        output_path (path): Selects base path for determining output path
        config (dataClass): Base class selecting user inputs
        verbose (bool): If True, prints out more information

    Outputs:
        An `.npz` file which includes:
        - The compressed data
        - The data headers
        - Normalization features if `config.apply_normalization=True`
    """

    verbose = config.verbose

    print("Compressing...")
    start = time.time()
    normalization_features = []

    if config.apply_normalization:
        normalization_features = normalization_features

    (
        compressed,
        error_bound_batch,
        error_bound_deltas,
        error_bound_index,
        original_shape,
    ) = helper.compress(
        model_path=model_output_path,
        config=config,
    )

    end = time.time()

    print("Compression took:", f"{(end - start) / 60:.3} minutes")

    names = np.load(config.input_path)["names"]

    # if config.extra_compression:
    #     if verbose:
    #         print("Extra compression selected")
    #         print(
    #             f"Saving compressed file to {os.path.join(output_path, 'compressed_output', 'compressed.npz')}"
    #         )
    #     np.savez_compressed(
    #         os.path.join(output_path, "compressed_output", "compressed.npz"),
    #         data=compressed,
    #         names=names,
    #         normalization_features=normalization_features,
    #     )
    # else:
    #     if verbose:
    #         print("Extra compression not selected")
    #         print(
    #             f"Saving compressed file to {os.path.join(output_path, 'compressed_output', 'compressed.npz')}"
    #         )
    #     np.savez(
    #         os.path.join(output_path, "compressed_output", "compressed.npz"),
    #         data=compressed,
    #         names=names,
    #         normalization_features=normalization_features,
    #     )

    # if config.save_error_bounded_deltas:
    #     error_bound_batch_index = np.array(
    #         [error_bound_batch, error_bound_index], dtype=object
    #     )
    #     f_batch_index = gzip.GzipFile(
    #         os.path.join(
    #             output_path,
    #             "compressed_output",
    #             "compressed_batch_index_metadata.npz.gz",
    #         ),
    #         "w",
    #     )
    #     f_deltas = gzip.GzipFile(
    #         os.path.join(output_path, "compressed_output", "compressed_deltas.npz.gz"),
    #         "w",
    #     )
    #     np.save(file=f_deltas, arr=error_bound_deltas)
    #     np.save(
    #         file=f_batch_index,
    #         arr=error_bound_batch_index,
    #     )
    #     f_batch_index.close()
    #     f_deltas.close()

    return (compressed, names, original_shape)



# Copyright 2022 Baler Contributors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn
from scipy.stats import wasserstein_distance
from torch.nn import functional
from tqdm.autonotebook import tqdm
from torch.nn import functional as F
from torch import distributions as dist

factor = 0.5
min_lr = 1e-6


def loss_function_swae(
    inputs,
    z,
    reconstructions,
    latent_dim,
    reg_weight=100,
    wasserstein_deg=2.0,
    num_projections=2000,
    projection_dist="normal",
):
    batch_size = inputs.shape[0]
    bias_corr = batch_size * (batch_size - 1)
    reg_weight = reg_weight / bias_corr

    mse_sum = nn.MSELoss(reduction="sum")
    mse_loss = mse_sum(reconstructions, inputs)
    number_of_columns = inputs.shape[1]
    mse_sum_loss = mse_loss / number_of_columns
    # recons_loss_l1 = F.l1_loss(reconstructions, inputs)
    recons_loss_l1 = 0
    swd_loss = compute_swd(
        z, wasserstein_deg, reg_weight, latent_dim, num_projections, projection_dist
    )
    loss = mse_sum_loss + recons_loss_l1 + swd_loss
    SWD = swd_loss
    return loss, mse_sum_loss, SWD


def compute_swd(z, p, reg_weight, latent_dim, num_projections, proj_dist):
    prior_z = torch.randn_like(z)  # [N x D]
    device = z.device

    proj_matrix = (
        get_random_projections(proj_dist, latent_dim, num_samples=num_projections)
        .transpose(0, 1)
        .to(device)
    )

    latent_projections = z.matmul(proj_matrix)  # [N x S]
    prior_projections = prior_z.matmul(proj_matrix)  # [N x S]

    # The Wasserstein distance is computed by sorting the two projections
    # across the batches and computing their element-wise distance
    w_dist = (
        torch.sort(latent_projections.t(), dim=1)[0]
        - torch.sort(prior_projections.t(), dim=1)[0]
    )
    w_dist = w_dist.pow(p)
    return reg_weight * w_dist.mean()


def get_random_projections(proj_dist, latent_dim, num_samples):
    if proj_dist == "normal":
        rand_samples = torch.randn(num_samples, latent_dim)
    elif proj_dist == "cauchy":
        rand_samples = (
            dist.Cauchy(torch.tensor([0.0]), torch.tensor([1.0]))
            .sample((num_samples, latent_dim))
            .squeeze()
        )
    else:
        raise ValueError("Unknown projection distribution.")

    rand_proj = rand_samples / rand_samples.norm(dim=1).view(-1, 1)
    return rand_proj  # [S x D]


def mse_loss_emd_l1(model_children, true_data, reconstructed_data, reg_param, validate):
    """
    Computes a sparse loss function consisting of three terms: the Earth Mover's Distance (EMD) loss between the
    true and reconstructed data, the mean squared error (MSE) loss between the reconstructed and true data, and a
    L1 regularization term on the output of a list of model children.
    Args: model_children (list): List of PyTorch modules representing the model architecture to be regularized.
    true_data (torch.Tensor): The ground truth data, with shape (batch_size, num_features). reconstructed_data (
    torch.Tensor): The reconstructed data, with shape (batch_size, num_features). reg_param (float): The weight of
    the L1 regularization term in the loss function. validate (bool): If True, returns only the EMD loss. If False,
    computes the full loss with the L1 regularization term.
    Returns:
        If validate is False, returns a tuple with three elements:
        - loss (torch.Tensor): The full sparse loss function, with shape ().
        - emd_loss (float): The EMD loss between the true and reconstructed data.
        - l1_loss (torch.Tensor): The L1 regularization term on the output of the model children.

        If validate is True, returns only the EMD loss as a float.
    """
    mse = nn.MSELoss()
    mse_loss = mse(reconstructed_data, true_data)
    wasserstein_distance_list = [
        wasserstein_distance(
            true_data.detach().numpy()[i, :], reconstructed_data.detach().numpy()[i, :]
        )
        for i in range(len(true_data))
    ]
    emd_loss = sum(wasserstein_distance_list)

    l1_loss = torch.Tensor(0)
    values = true_data
    if not validate:
        for i in range(len(model_children)):
            values = model_children[i](values)
            l1_loss += torch.mean(torch.abs(values))

        loss = emd_loss + mse_loss + reg_param * l1_loss
        return loss, emd_loss, l1_loss
    else:
        return emd_loss


def mse_loss_l1(model_children, true_data, reconstructed_data, reg_param, validate):
    # This function is a modified version of the original function by George Dialektakis found at
    # https://github.com/Autoencoders-compression-anomaly/Deep-Autoencoders-Data-Compression-GSoC-2021
    # Released under the Apache License 2.0 found at https://www.apache.org/licenses/LICENSE-2.0.txt
    # Copyright 2021 George Dialektakis

    """
    Computes a sparse loss function consisting of two terms: the mean squared error (MSE) loss between the
    reconstructed and true data, and a L1 regularization term on the output of a list of model children.
    Args: model_children (list): List of PyTorch modules representing the model architecture to be regularized.
    true_data (torch.Tensor): The ground truth data, with shape (batch_size, num_features). reconstructed_data (
    torch.Tensor): The reconstructed data, with shape (batch_size, num_features). reg_param (float): The weight of
    the L1 regularization term in the loss function. validate (bool): If True, returns only the MSE loss. If False,
    computes the full loss with the L1 regularization term.
    Returns:
        If validate is False, returns a tuple with three elements:
        - loss (torch.Tensor): The full sparse loss function, with shape ().
        - mse_loss (float): The MSE loss between the true and reconstructed data.
        - l1_loss (torch.Tensor): The L1 regularization term on the output of the model children.

        If validate is True, returns a tuple with three elements:
        - mse_loss (torch.Tensor): The MSE loss between the true and reconstructed data.
        - 0.
        - 0.
    """
    mse = nn.MSELoss()
    mse_loss = mse(reconstructed_data, true_data)

    l1_loss = 0
    values = true_data
    if not validate:
        for i in range(len(model_children)):
            values = functional.relu(model_children[i](values))
            l1_loss += torch.mean(torch.abs(values))

        loss = mse_loss + reg_param * l1_loss
        return loss, mse_loss, l1_loss
    else:
        return mse_loss, 0, 0


def mse_sum_loss_l1(model_children, true_data, reconstructed_data, reg_param, validate):
    """
    Computes the sum of mean squared error (MSE) loss and L1 regularization loss.
    Args:
        model_children (list): List of PyTorch modules representing the encoder network.
        true_data (tensor): Ground truth tensor of shape (batch_size, input_size).
        reconstructed_data (tensor): Reconstructed tensor of shape (batch_size, input_size).
        reg_param (float): Regularization parameter for L1 loss.
        validate (bool): Whether to return only MSE loss or both MSE and L1 losses.
    Returns:
        If validate is False:
            loss (tensor): Total loss consisting of MSE loss and L1 regularization loss.
            mse_sum_loss (tensor): Mean squared error loss.
            l1_loss (tensor): L1 regularization loss.
        If validate is True:
            mse_sum_loss (tensor): Mean squared error loss.
            0 (int): Placeholder for MSE loss since it is not calculated during validation.
            0 (int): Placeholder for L1 loss since it is not calculated during validation.
    """
    mse_sum = nn.MSELoss(reduction="sum")
    mse_loss = mse_sum(reconstructed_data, true_data)
    number_of_columns = true_data.shape[1]

    mse_sum_loss = mse_loss / number_of_columns

    l1_loss = 0
    values = true_data
    if not validate:
        for i in range(len(model_children)):
            values = functional.relu(model_children[i](values))
            l1_loss += torch.mean(torch.abs(values))

        loss = mse_sum_loss + reg_param * l1_loss
        return loss, mse_sum_loss, l1_loss
    else:
        return mse_sum_loss, 0, 0


# Accuracy function still WIP. Not working properly.
# Probably has to do with total_correct counter.


def accuracy(model, dataloader):
    """
    Computes the accuracy of a PyTorch model on a given dataset.
    Args:
        model (nn.Module): The PyTorch model to evaluate.
        dataloader (DataLoader): DataLoader object containing the dataset to evaluate on.
    Returns:
        accuracy_frac (float): The fraction of correctly classified instances in the dataset.
    """
    print("Accuracy")
    model.eval()

    total_correct = 0
    total_instances = 0

    with torch.no_grad():
        for data in tqdm(dataloader):
            x, _ = data
            classifications = torch.argmax(x)

            correct_pred = torch.sum(classifications == x).item()

            total_correct += correct_pred
            total_instances += len(x)

    accuracy_frac = round(total_correct / total_instances, 3)
    print(accuracy_frac)
    return accuracy_frac


class EarlyStopping:
    """
    Class to perform early stopping during model training.
    Attributes:
        patience (int): The number of epochs to wait before stopping the training process if the
            validation loss doesn't improve.
        min_delta (float): The minimum difference between the new loss and the previous best loss
            for the new loss to be considered an improvement.
        counter (int): Counts the number of times the validation loss hasn't improved.
        best_loss (float): The best validation loss observed so far.
        early_stop (bool): Flag that indicates whether early stopping criteria have been met.
    """

    def __init__(self, patience: int, min_delta: float):
        self.patience = patience  # Nr of times we allow val. loss to not improve before early stopping
        self.min_delta = min_delta  # min(new loss - best loss) for new loss to be considered improvement
        self.counter = 0  # counts nr of times val_loss dosent improve
        self.best_loss = None
        self.early_stop = False

    def __call__(self, train_loss):
        if self.best_loss is None:
            self.best_loss = train_loss

        elif self.best_loss - train_loss > self.min_delta:
            self.best_loss = train_loss
            self.counter = 0  # Resets if val_loss improves

        elif self.best_loss - train_loss < self.min_delta:
            self.counter += 1

            print(f"Early stopping counter {self.counter} of {self.patience}")
            if self.counter >= self.patience:
                print("Early Stopping")
                self.early_stop = True


class LRScheduler:
    """
    A learning rate scheduler that adjusts the learning rate of an optimizer based on the training loss.
    Args:
        optimizer (torch.optim.Optimizer): The optimizer whose learning rate will be adjusted.
        patience (int): The number of epochs with no improvement in training loss after which the learning rate
            will be reduced.
        min_lr (float, optional): The minimum learning rate that can be reached (default: 1e-6).
        factor (float, optional): The factor by which the learning rate will be reduced (default: 0.1).
    Attributes:
        lr_scheduler (torch.optim.lr_scheduler.ReduceLROnPlateau): The PyTorch learning rate scheduler that
            actually performs the adjustments.
    Example usage:
        optimizer = torch.optim.Adam(model.parameters(), lr=0.01)
        lr_scheduler = LRScheduler(optimizer, patience=3, min_lr=1e-5, factor=0.5)
        for epoch in range(num_epochs):
            train_loss = train(model, train_data_loader)
            lr_scheduler(train_loss)
            # ...
    """

    def __init__(self, optimizer, patience, min_lr=min_lr, factor=factor):
        self.optimizer = optimizer
        self.patience = patience
        self.min_lr = min_lr
        self.factor = factor

        # Maybe add if statements for selectment of lr schedulers
        self.lr_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            self.optimizer,
            mode="min",
            patience=self.patience,
            factor=self.factor,
            min_lr=self.min_lr,
            verbose=True,
        )

    def __call__(self, train_loss):
        self.lr_scheduler.step(train_loss)



# Copyright 2022 Baler Contributors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import numpy as np
import pytest
import torch
from sklearn.preprocessing import MinMaxScaler

from baler.modules import data_processing
from baler.modules import helper


def test_import_config_success():
    # Call the import_config function with the sample config file path
    config = helper.Config
    config.Foo = "Bar"
    config.Baz = 10

    # Assert that the result is equal to the expected config
    # This checks that the import_config function correctly loads the JSON file and returns the expected dictionary
    assert config.Foo == "Bar"


def test_save_model():
    # Test data
    model = torch.nn.Linear(3, 2)
    model_path = "test_model.pt"

    # Save the model
    data_processing.save_model(model, model_path)

    # Check that the model file has been created
    assert os.path.exists(model_path)

    # Clean up
    os.remove(model_path)


@pytest.fixture
def minmax_test_data():
    return [
        (np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), np.array([[1, 2, 3], [6, 6, 6]])),
        (
            np.array([[-1, -2, -3], [-4, -5, -6], [-7, -8, -9]]),
            np.array([[-7, -8, -9], [6, 6, 6]]),
        ),
        (np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]]), np.array([[0, 0, 0], [2, 2, 2]])),
    ]


def test_find_minmax_success(minmax_test_data):
    for data, expected_result in minmax_test_data:
        result = data_processing.find_minmax(data)
        assert np.array_equal(result, expected_result)


def test_normalize():
    # Test data
    data = [1, 2, 3, 4, 5]

    # Test configuration 1
    custom_norm1 = False
    expected_result1 = np.array([0.0, 0.25, 0.5, 0.75, 1.0])

    # Test configuration 2
    custom_norm2 = True
    expected_result2 = np.array([1, 2, 3, 4, 5])

    # Test the normalize function with the test data and configuration 1
    result1 = data_processing.normalize(data, custom_norm1)
    np.testing.assert_almost_equal(result1, expected_result1)

    # Test the normalize function with the test data and configuration 2
    result2 = data_processing.normalize(data, custom_norm2)
    np.testing.assert_almost_equal(result2, expected_result2)


def test_renormalize_std():
    # Test data
    data = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
    true_min = 1
    feature_range = 2

    # Renormalize the data using the renormalize_std function
    renormalized_data = data_processing.renormalize_std(data, true_min, feature_range)

    # Check that the renormalized data is correct
    expected_renormalized_data = np.array([1.2, 1.4, 1.6, 1.8, 2.0])
    np.testing.assert_array_equal(renormalized_data, expected_renormalized_data)


def test_renormalize_func():
    # Test data
    scaler = MinMaxScaler()
    data = [[-1, 2], [-0.5, 6], [0, 10], [1, 18]]
    scaler.fit(data)
    norm_data = scaler.transform(data)
    true_min = [-1, 2]
    feature_range = [2, 16]

    # Renormalize the data using the renormalize_std function
    renormalized_data = data_processing.renormalize_func(
        norm_data, true_min, feature_range
    )

    # Check that the renormalized data is correct
    expected_renormalized_data = np.array([[-1, 2], [-0.5, 6], [0, 10], [1, 18]])
    np.testing.assert_array_equal(renormalized_data, expected_renormalized_data)



# Copyright 2022 Baler Contributors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import importlib
import os
import sys
from dataclasses import dataclass
from math import ceil
import gzip

from tqdm.autonotebook import tqdm

sys.path.append(os.getcwd())
import numpy as np
import torch
from torch.utils.data import DataLoader
from sklearn.model_selection import train_test_split

from baler_compressor import data_processing
from numba import jit


def get_arguments():
    """Determines the arguments one is able to apply in the command line when running Baler. Use `--help` to see what
    options are available.

    Returns: .py, string, folder: `.py` file containing the config options, string determining what mode to run,
    projects directory where outputs go.
    """
    parser = argparse.ArgumentParser(
        prog="baler",
        description=(
            "Baler is a machine learning based compression tool for big data.\n\n"
            "Baler has three running modes:\n\n"
            '\t1. Derivation: Using a configuration file and a "small" input dataset, Baler derives a '
            "machine learning model optimized to compress and decompress your data.\n\n"
            "\t2. Compression: Using a previously derived model and a large input dataset, Baler compresses "
            "your data and outputs a smaller compressed file.\n\n"
            "\t3. Decompression: Using a previously compressed file as input and a model, Baler decompresses "
            "your data into a larger file."
        ),
        epilog="Enjoy!",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--mode",
        type=str,
        required=True,
        help="newProject, train, compress, decompress, plot, info",
    )
    parser.add_argument(
        "--project",
        type=str,
        required=True,
        nargs=2,
        metavar=("WORKSPACE", "PROJECT"),
        help="Specifies workspace and project.\n"
        "e.g. --project CFD firstTry"
        ", specifies workspace 'CFD' and project 'firstTry'\n\n"
        "When combined with newProject mode:\n"
        "  1. If workspace and project exist, take no action.\n"
        "  2. If workspace exists but project does not, create project in workspace.\n"
        "  3. If workspace does not exist, create workspace directory and project.",
    )
    parser.add_argument(
        "--pytorchProfile", action="store_true", help="Enable PyTorch profiling"
    )
    parser.add_argument(
        "--energyProfile", action="store_true", help="Enable Energy profiling"
    )
    parser.add_argument(
        "--verbose", dest="verbose", action="store_true", help="Verbose mode"
    )
    parser.set_defaults(verbose=False)

    args = parser.parse_args()

    workspace_name = args.project[0]
    project_name = args.project[1]
    config_path = (
        f"workspaces.{workspace_name}.{project_name}.config.{project_name}_config"
    )

    if args.mode == "newProject":
        config = None
    else:
        config = Config
        importlib.import_module(config_path).set_config(config)

    return (
        config,
        args.mode,
        workspace_name,
        project_name,
        args.verbose,
        args.pytorchProfile,
        args.energyProfile,
    )


def create_new_project(
    workspace_name: str,
    project_name: str,
    verbose: bool = False,
    base_path: str = "workspaces",
) -> None:
    """Creates a new project directory output subdirectories and config files within a workspace.

    Args:
        workspace_name (str): Creates a workspace (dir) for storing data and projects with this name.
        project_name (str): Creates a project (dir) for storing configs and outputs with this name.
        verbose (bool, optional): Whether to print out the progress. Defaults to False.
    """

    # Create full project path
    workspace_path = os.path.join(base_path, workspace_name)
    project_path = os.path.join(base_path, workspace_name, project_name)
    if os.path.exists(project_path):
        print(f"The workspace and project ({project_path}) already exists.")
        return
    os.makedirs(project_path)

    # Create required directories
    required_directories = [
        os.path.join(workspace_path, "data"),
        os.path.join(project_path, "config"),
        os.path.join(project_path, "output", "compressed_output"),
        os.path.join(project_path, "output", "decompressed_output"),
        os.path.join(project_path, "output", "plotting"),
        os.path.join(project_path, "output", "training"),
    ]

    if verbose:
        print(f"Creating project {project_name} in workspace {workspace_name}...")
    for directory in required_directories:
        if verbose:
            print(f"Creating directory {directory}...")
        os.makedirs(directory, exist_ok=True)

    # Populate default config
    with open(
        os.path.join(project_path, "config", f"{project_name}_config.py"), "w"
    ) as f:
        f.write(create_default_config(workspace_name, project_name))


@dataclass
class Config:
    """Defines a configuration dataclass"""

    input_path: str
    compression_ratio: float
    epochs: int
    early_stopping: bool
    early_stoppin_patience: int
    lr_scheduler: bool
    lr_scheduler_patience: int
    min_delta: int
    model_name: str
    model_type = str
    custom_norm: bool
    l1: bool
    reg_param: float
    RHO: float
    lr: float
    batch_size: int
    test_size: float
    data_dimension: int
    intermittent_model_saving: bool
    intermittent_saving_patience: int
    mse_avg: bool
    mse_sum: bool
    emd: bool
    l1: bool
    deterministic_algorithm: bool


def create_default_config(workspace_name: str, project_name: str) -> str:
    """Creates a default config file for a project.
    Args:
        workspace_name (str): Name of the workspace.
        project_name (str): Name of the project.
    Returns:
        str: Default config file.
    """

    return f"""
# === Configuration options ===

def set_config(c):
    c.input_path                   = "workspaces/{workspace_name}/data/{project_name}_data.npz"
    c.data_dimension               = 1
    c.compression_ratio            = 2.0
    c.apply_normalization          = True
    c.model_name                   = "AE"
    c.model_type                    = "dense"
    c.epochs                       = 5
    c.lr                           = 0.001
    c.batch_size                   = 512
    c.early_stopping               = True
    c.lr_scheduler                 = True




# === Additional configuration options ===

    c.early_stopping_patience      = 100
    c.min_delta                    = 0
    c.lr_scheduler_patience        = 50
    c.custom_norm                  = False
    c.reg_param                    = 0.001
    c.RHO                          = 0.05
    c.test_size                    = 0
    # c.number_of_columns            = 24
    # c.latent_space_size            = 12
    c.extra_compression            = False
    c.intermittent_model_saving    = False
    c.intermittent_saving_patience = 100
    c.mse_avg                      = False
    c.mse_sum                      = True
    c.emd                          = False
    c.l1                           = True
    c.activation_extraction        = False
    c.deterministic_algorithm      = True

"""


def model_init(model_name: str):
    """Calls `data_processing.initialise_model`.

    Args:
        model_name (string): The name of the model you wish to initialize

    Returns:
       nn.Module : The initialized model
    """
    # This is used when we don't have saved model parameters.
    model_object = data_processing.initialise_model(model_name)
    return model_object


def numpy_to_tensor(data):
    """Converts ndarray to torch.Tensors.

    Args:
        data (ndarray): The data you wish to convert from ndarray to torch.Tensor.

    Returns:
        torch.Tensor: Your data as a tensor
    """
    return torch.from_numpy(data)


@jit(cache=True, parallel=True, forceobj=True)
def normalize(data, custom_norm):
    """Applies `data_processing.normalize()` along every axis of given data

    Args:
        data (ndarray): Data you wish to normalize
        custom_norm (boolean): Wether or not you wish to use MinMax normalization

    Returns:
        ndarray: Normalized data
    """
    data = np.apply_along_axis(
        data_processing.normalize, axis=0, arr=data, custom_norm=custom_norm
    )
    return data


@jit(cache=True, parallel=True, forceobj=True)
def process(
    input_path,
    custom_norm,
    test_size,
    apply_normalization,
    convert_to_blocks,
    verbose,
):
    """Loads the input data into a ndarray, splits it into train/test splits and normalizes if chosen.

    Args:
        input_path (string): Path to input data
        custom_norm (boolean): `False` if you want to use MinMax normalization. `True` otherwise.
        test_size (float): How much of your data to use as validation.
        apply_normalization (boolean): `True` if you want to normalize. `False` if you don't want to normalize.

    Returns: ndarray, ndarray, ndarray: Array with the train set, array with the test set and array with the
    normalization features.
    """
    loaded = np.load(input_path)
    data = loaded["data"]

    if verbose:
        print("Original Dataset Shape - ", data.shape)

    original_shape = data.shape

    if convert_to_blocks:
        data = data_processing.convert_to_blocks_util(convert_to_blocks, data)

    normalization_features = data_processing.find_minmax(data)
    if apply_normalization:
        print("Normalizing the data...")
        data = normalize(data, custom_norm)
    if not test_size:
        train_set = data
        test_set = train_set
    else:
        train_set, test_set = train_test_split(
            data, test_size=test_size, random_state=1
        )

    return (train_set, test_set, normalization_features, original_shape)


def renormalize(data, true_min_list, feature_range_list):
    """Calls `data_processing.renormalize_func()`.

    Args:
        data (ndarray): Data you wish to un-normalize
        true_min_list (ndarray): List of column minimums
        feature_range_list (ndarray): List of column feature ranges

    Returns:
        ndarray: Un-normalized array
    """
    return data_processing.renormalize_func(data, true_min_list, feature_range_list)



def plotter(output_path, config):
    """Calls `plotting.plot()`

    Args:
        output_path (string): Path to the output directory
        config (dataClass): Base class selecting user inputs

    """

    plotting.plot(output_path, config)
    print("=== Done ===")
    print("Your plots are available in:", os.path.join(output_path, "plotting"))


def loss_plotter(path_to_loss_data, output_path, config):
    """Calls `plotting.loss_plot()`

    Args:
        path_to_loss_data (string): Path to the values for the loss plot
        output_path (string): Path to output the data
        config (dataClass): Base class selecting user inputs

    Returns:
        .pdf file: Plot containing the loss curves
    """
    return plotting.loss_plot(path_to_loss_data, output_path, config)


def model_saver(model, model_path):
    """Calls `data_processing.save_model()`

    Args:
        model (nn.Module): The PyTorch model to save.
        model_path (str): String defining the models save path.

    Returns:
        .pt file: `.pt` File containing the model state dictionary
    """
    return data_processing.save_model(model, model_path)


def detacher(tensor):
    """Detaches a given tensor to ndarray

    Args:
        tensor (torch.Tensor): The PyTorch tensor one wants to convert to a ndarray

    Returns:
        ndarray: Converted torch.Tensor to ndarray
    """
    return tensor.cpu().detach().numpy()


def get_device():
    """Returns the appropriate processing device. IF cuda is available it returns "cuda:0"
        Otherwise it returns "cpu"

    Returns:
        _type_: Device string, either "cpu" or "cuda:0"
    """
    device = None
    if torch.cuda.is_available():
        dev = "cuda:0"
        device = torch.device(dev)
    else:
        dev = "cpu"
        device = torch.device(dev)
    return device


@jit(cache=True, parallel=True, forceobj=True)
def save_error_bounded_requirement(config, decoded_output, data_batch):
    rms_pred_error = (
        np.divide(
            np.subtract(decoded_output, data_batch),
            data_batch,
        )
        * 100
    )

    # Ignoring RMS Undefind Values because Ground Truth is Zero
    rms_pred_error[
        (rms_pred_error == np.inf)
        | (rms_pred_error == -np.inf)
        | (rms_pred_error == np.nan)
    ] = 0.0
    rms_pred_error_index = np.where(
        abs(rms_pred_error) > config.error_bounded_requirement
    )
    rows_idx, col_idx = rms_pred_error_index
    if len(rows_idx) > 0 and len(col_idx) > 0:
        rms_pred_error_exceeding_error_bound = np.subtract(
            decoded_output,
            data_batch,
            dtype=np.float16,
        )
        deltas = []
        for i in range(len(rows_idx)):
            deltas.append(rms_pred_error_exceeding_error_bound[rows_idx[i]][col_idx[i]])
    return deltas, rms_pred_error_index


def compress(model_path, config):
    """Function which performs the compression of the input file. In order to compress, you must have a dataset whose
    path is determined by `input_path` in the `config`. You also need a trained model from path `model_path`. The
    model path is then used to initialize the model used for compression. The data is then converted into a
    `torch.tensor` and then passed through `model.encode`.

    Args:
        model_path (string): Path to where the model is located
        config (dataClass): Base class selecting user inputs

    Raises:
        NameError: Baler currently only supports 1D (e.g. HEP) or 2D (e.g. CFD) data as inputs.

    Returns:
        torch.Tensor: Compressed data as PyTorch tensor
    """

    # Loads the data and applies normalization if config.apply_normalization = True
    loaded = np.load(config.input_path)
    data_before = loaded["data"]
    original_shape = data_before.shape

    if hasattr(config, "convert_to_blocks") and config.convert_to_blocks:
        data_before = data_processing.convert_to_blocks_util(
            config.convert_to_blocks, data_before
        )

    if config.apply_normalization:
        print("Normalizing...")
        data = normalize(data_before, config.custom_norm)
    else:
        data = data_before
    number_of_columns = 0
    try:
        n_features = 0
        if config.data_dimension == 1:
            column_names = np.load(config.input_path)["names"]
            number_of_columns = len(column_names)
            config.latent_space_size = ceil(
                number_of_columns / config.compression_ratio
            )
            config.number_of_columns = number_of_columns
            n_features = number_of_columns
        elif config.data_dimension == 2:
            if config.model_type == "dense":
                number_of_rows = data.shape[1]
                config.number_of_columns = data.shape[2]
                n_features = number_of_rows * config.number_of_columns
            else:
                number_of_rows = original_shape[1]
                config.number_of_columns = original_shape[2]
                n_features = config.number_of_columns
            config.latent_space_size = ceil(
                (number_of_rows * config.number_of_columns) / config.compression_ratio
            )
        else:
            raise NameError(
                "Data dimension can only be 1 or 2. Got config.data_dimension = "
                + str(config.data_dimension)
            )
    except AttributeError:
        number_of_columns = config.number_of_columns
        latent_space_size = config.latent_space_size
        print(f"{number_of_columns} -> {latent_space_size} dimensions")

    # Initialise and load the model correctly.
    latent_space_size = config.latent_space_size
    bs = config.batch_size
    device = get_device()
    model_object = data_processing.initialise_model(config.model_name)
    model = data_processing.load_model(
        model_object,
        model_path=model_path,
        n_features=n_features,
        z_dim=config.latent_space_size,
    )
    model.eval()

    if config.data_dimension == 2:
        if config.model_type == "convolutional" and config.model_name == "Conv_AE_3D":
            data_tensor = torch.tensor(data, dtype=torch.float32).view(
                data.shape[0] // bs, 1, bs, data.shape[1], data.shape[2]
            )
        elif config.model_type == "convolutional":
            data_tensor = torch.tensor(data, dtype=torch.float32).view(
                data.shape[0], 1, data.shape[1], data.shape[2]
            )
        elif config.model_type == "dense":
            data_tensor = torch.tensor(data, dtype=torch.float32).view(
                data.shape[0], data.shape[1] * data.shape[2]
            )
    elif config.data_dimension == 1:
        data_tensor = torch.tensor(data, dtype=torch.float64)

    # Batching data to avoid memory leaks
    data_dl = DataLoader(
        data_tensor,
        batch_size=bs,
        shuffle=False,
        drop_last=False,
    )

    # Perform compression
    error_bound_batch, error_bound_deltas, error_bound_index, compressed = (
        [],
        [],
        [],
        [],
    )

    with torch.no_grad():
        for idx, data_batch in enumerate(tqdm(data_dl)):
            data_batch = data_batch.to(device)

            compressed_output = model.encode(data_batch)

#            if config.save_error_bounded_deltas:
#                decoded_output = model.decode(compressed_output)
#                decoded_output = detacher(decoded_output)
#                deltas_compressed = 0
            # Converting back to numpyarray
            compressed_output = detacher(compressed_output)
            data_batch = detacher(data_batch)

#            if config.save_error_bounded_deltas:
#                (
#                    deltas,
#                    rms_pred_error_index,
#                ) = save_error_bounded_requirement(config, decoded_output, data_batch)
#                if len(rms_pred_error_index) > 0:
#                    error_bound_batch.append(idx)
#                    error_bound_deltas.append(deltas)
#                    error_bound_index.append(rms_pred_error_index)
#                    deltas_compressed += len(rms_pred_error_index[0])

            if idx == 0:
                compressed = compressed_output
            else:
                compressed = np.concatenate((compressed, compressed_output))

#    if config.save_error_bounded_deltas:
#        print("Total Deltas Found - ", deltas_compressed)

    return (
        compressed,
        error_bound_batch,
        error_bound_deltas,
        error_bound_index,
        original_shape,
    )


def decompress(
    model_path,
    input_path,
    input_path_deltas,
    input_batch_index,
    model_name,
    config,
    output_path,
):
    """Function which performs the decompression of the compressed file. In order to decompress, you must have a
    compressed file, whose path is determined by `input_path`, a model from path `model_path` and a model_name. The
    model path and model names are used to initialize the model used for decompression. The data is then converted
    into a `torch.tensor` and then passed through `model.decode`.

    Args:
        model_path (string): Path to where the model is located
        input_path (string): Path to the data you want to decompress
        model_name (string): Name of trained model from which you want to use decode
        config (dataClass): Base class selecting user inputs

    Returns: torch.tensor, ndarray, ndarray: decompressed data as tensor, ndarray of column names, ndarray of
    normalization features
    """

    # Load the data & define necessary variables
    loaded = np.load(input_path)
    data = loaded["data"]
    names = loaded["names"]
    original_shape = loaded["original_shape"]
    normalization_features = loaded["normalization_features"]

    if config.model_type == "convolutional":
        final_layer_details = np.load(
            os.path.join(output_path, "training", "final_layer.npy"),
        )

#    if config.save_error_bounded_deltas:
#        loaded_deltas = np.load(
#            gzip.GzipFile(input_path_deltas, "r"), allow_pickle=True
#        )
#        loaded_batch_indexes = np.load(
#            gzip.GzipFile(input_batch_index, "r"), allow_pickle=True
#        )
#        error_bound_batch = loaded_batch_indexes[0]
#        error_bound_deltas = loaded_deltas
#        error_bound_index = loaded_batch_indexes[1]
#        deltas_added = 0

    model_name = config.model_name
    latent_space_size = len(data[0])
    bs = config.batch_size
    model_dict = torch.load(str(model_path), map_location=get_device())
    if config.data_dimension == 2 and config.model_type == "dense":
        number_of_columns = int((len(model_dict[list(model_dict.keys())[-1]])))
    else:
        number_of_columns = len(model_dict[list(model_dict.keys())[-1]])

    # Initialise and load the model correctly.
    device = get_device()
    model_object = data_processing.initialise_model(model_name)
    model = data_processing.load_model(
        model_object,
        model_path=model_path,
        n_features=number_of_columns,
        z_dim=latent_space_size,
    )
    model.eval()

    if config.model_type == "convolutional":
        model.set_final_layer_dims(final_layer_details)

    # Load the data, convert to tensor and batch it to avoid memory leaks
    data_tensor = torch.from_numpy(data).to(device)
    data_dl = DataLoader(
        data_tensor,
        batch_size=bs,
        shuffle=False,
        drop_last=False,
    )

    # Perform Decompression
    decompressed = []
    with torch.no_grad():
        for idx, data_batch in enumerate(tqdm(data_dl)):
            data_batch = data_batch.to(device)

            out = model.decode(data_batch).to(device)
            # Converting back to numpyarray
            out = detacher(out)
#            if config.save_error_bounded_deltas:
#                if idx in error_bound_batch:
#                    # Error Bounded Deltas added to Decompressed output
#                    delta_idx = np.where(error_bound_batch == idx)
#                    deltas = error_bound_deltas[delta_idx][0]
#                    delta_index = error_bound_index[delta_idx][0]
#                    row_idx, col_idx = delta_index

#                    for i in range(len(row_idx)):
#                        out[row_idx[i]][col_idx[i]] -= deltas[i]
#                        deltas_added += 1

            if idx == 0:
                decompressed = out
            else:
                decompressed = np.concatenate((decompressed, out))

#    if config.save_error_bounded_deltas:
#        print("Total Deltas Added - ", deltas_added)

    if config.data_dimension == 2 and config.model_type == "dense":
        decompressed = decompressed.reshape(
            (len(decompressed), original_shape[1], original_shape[2])
        )

    return decompressed, names, normalization_features, original_shape


def diagnose(input_path: str, output_path: str) -> None:
    """Calls diagnostics.diagnose()

    Args:
        input_path (str): path to the np.array contataining the activations values
        output_path (str): path to store the diagnostics pdf
    """
    diagnostics.diagnose(input_path, output_path)


def perform_hls4ml_conversion(output_path, config):
    """Function which performs the conversion of the model to FPGA architecture using hls4ml. In order to convert, you must have a trained model.

    The output hls4ml project will be located in OutputDir defined in the config file.
    The model path is determined by the `output_path`. The 'output_path' should point to the output directory inside the project directory.

    Before running this function:
    1. Install Vivado 2020.1 and add in to path.
    2. Install hls4ml and tensorflow to the projects virtual environment.
    3. Add the following configuration options to the projects config:

        c.default_reuse_factor
        c.default_precision
        c.Strategy
        c.Part
        c.ClockPeriod
        c.IOType
        c.InputShape
        c.ProjectName
        c.OutputDir
        c.InputData
        c.OutputPredictions
        c.csim
        c.synth
        c.cosim
        c.export

        This function was tested with the following configuration values:

        c.default_reuse_factor         = 1
        c.default_precision            = "ap_fixed<16,8>"
        c.Strategy                     = "latency"
        c.Part                         = "xcvu9p-flga2104-2L-e"
        c.ClockPeriod                  = 5
        c.IOType                       = "io_parallel"
        c.InputShape                   = (1,16)
        c.ProjectName                  = "tiny_test_model"
        c.OutputDir                    = "workspaces/FPGA_compression_workspace/first_FPGA_Compression_project/output/hls4ml"
        c.InputData                    = None
        c.OutputPredictions            = None
        c.csim                         = False
        c.synth                        = True
        c.cosim                        = False
        c.export                       = False

        Naming is based on the names used by hls4ml for easier reference. For more details please see hls4ml documentation.

    4. The model to transform should be defined similarly to FPGA_prototype_model. Meaning:
        a. Activations should be defined as layers and not functions.
        b. If the model is defined as a class, each layer should be defined with an attribute. An attribute for the whole model instead of for each layer separately might cause errors in hls4ml compile function.
        c. Not all types of layers are supported in hls4ml, check hls4ml for the supported layers.



    Args:
    output_path (string): Path to the output directory inside the project directory.
    config (dataClass): Base class selecting user inputs

    Returns: None


    """

    import hls4ml

    model_path = os.path.join(output_path, "compressed_output", "model.pt")

    model_object = data_processing.initialise_model(config.model_name)
    model = data_processing.load_model(
        model_object,
        model_path=model_path,
        n_features=config.number_of_columns,
        z_dim=config.latent_space_size,
    )
    model.to("cpu")

    hls_config = hls4ml.utils.config_from_pytorch_model(
        model,
        granularity="name",
        default_reuse_factor=config.default_reuse_factor,
        default_precision=config.default_precision,
    )

    hls_config["Model"]["Strategy"] = config.Strategy

    cfg = hls4ml.converters.create_config(backend="Vivado")
    cfg["Part"] = config.Part
    cfg["ClockPeriod"] = config.ClockPeriod
    cfg["IOType"] = config.IOType
    cfg["HLSConfig"] = hls_config
    cfg["PytorchModel"] = model
    cfg["InputShape"] = config.InputShape
    cfg["OutputDir"] = config.OutputDir

    if config.InputData and config.OutputPredictions:
        cfg["InputData"] = config.InputData
        cfg["OutputPredictions"] = config.OutputPredictions

    hls_model = hls4ml.converters.pytorch_to_hls(cfg)
    hls_model.config.config["ProjectName"] = config.ProjectName

    hls_model.compile()

    hls_model.build(
        csim=config.csim, synth=config.synth, cosim=config.cosim, export=config.export
    )



import torch
from torch.utils.data import DataLoader, TensorDataset

from baler.modules.utils import (
    mse_loss_emd_l1,
    mse_loss_l1,
    accuracy,
    LRScheduler,
)


def test_mse_loss_emd_l1():
    # Generate random data
    batch_size, num_features = 32, 64
    true_data = torch.randn(batch_size, num_features)
    reconstructed_data = torch.randn(batch_size, num_features)

    # Create model children list
    model_children = [torch.nn.Linear(num_features, num_features)]

    # Compute loss with L1 regularization
    reg_param = 0.1
    loss, emd_loss, l1_loss = mse_loss_emd_l1(
        model_children, true_data, reconstructed_data, reg_param, validate=False
    )
    assert isinstance(loss, torch.Tensor)
    assert isinstance(emd_loss, float)
    assert isinstance(l1_loss, torch.Tensor)

    # Compute loss without L1 regularization
    loss = mse_loss_emd_l1(
        model_children, true_data, reconstructed_data, 0.0, validate=False
    )[0]
    assert isinstance(loss, torch.Tensor)

    # Compute EMD loss only
    emd_loss = mse_loss_emd_l1(
        model_children, true_data, reconstructed_data, 0.0, validate=True
    )
    assert isinstance(emd_loss, float)


def test_mse_loss_l1():
    model_children = [torch.nn.Linear(10, 10)]
    true_data = torch.randn(32, 10)
    reconstructed_data = torch.randn(32, 10)
    reg_param = 0.1
    validate = False

    loss, mse_loss, l1_loss = mse_loss_l1(
        model_children, true_data, reconstructed_data, reg_param, validate
    )

    assert isinstance(loss, torch.Tensor)
    assert isinstance(mse_loss, torch.Tensor)
    assert isinstance(l1_loss, torch.Tensor)
    assert loss.shape == torch.Size([])
    assert mse_loss.shape == torch.Size([])
    assert l1_loss.shape == torch.Size([])

    validate = True
    mse_loss, _, _ = mse_loss_l1(
        model_children, true_data, reconstructed_data, reg_param, validate
    )

    assert isinstance(mse_loss, torch.Tensor)
    assert mse_loss.shape == torch.Size([])


def test_accuracy():
    model = torch.nn.Sequential(
        torch.nn.Linear(10, 10), torch.nn.ReLU(), torch.nn.Linear(10, 2)
    )
    x = torch.randn(100, 10)
    y = torch.randint(0, 2, (100,))
    dataset = TensorDataset(x, y)
    dataloader = DataLoader(dataset, batch_size=10)
    accuracy_frac = accuracy(model, dataloader)
    assert isinstance(accuracy_frac, float)
    assert 0 <= accuracy_frac <= 1


def test_lr_scheduler():
    # Create a dummy model and optimizer
    model = torch.nn.Linear(10, 1)
    optimizer = torch.optim.SGD(model.parameters(), lr=0.1)

    # Create a learning rate scheduler
    lr_scheduler = LRScheduler(optimizer, patience=2, min_lr=1e-5, factor=0.5)

    # Test that the learning rate is not reduced when the loss is improving
    train_losses = [10.0, 9.0, 8.0, 7.0]
    for train_loss in train_losses:
        lr_scheduler(train_loss)
        assert optimizer.param_groups[0]["lr"] == 0.1

    # Test that the learning rate is reduced when the loss is not improving
    train_losses = [10.0, 9.0, 10.0, 11.0, 12.0]
    for i, train_loss in enumerate(train_losses):
        lr_scheduler(train_loss)
        if i >= 2:
            assert optimizer.param_groups[0]["lr"] == 0.05

    # Test that the learning rate does not go below the minimum value
    train_losses = [10.0] * 100
    for train_loss in train_losses:
        lr_scheduler(train_loss)
    assert optimizer.param_groups[0]["lr"] == 1e-5



import sys
sys.path.append("../src/")

# import baler
import baler_compressor.config as config_module
import baler_compressor.trainer as trainer_module
import baler_compressor.compressor as compressor_module
import baler_compressor.decompressor as decompressor_module
import baler_compressor.helper as baler_helper

# import helper for plotting
import helper

# import others
import torch
import numpy as np
import imageio.v2 as imageio
import matplotlib.pyplot as plt
from tqdm import tqdm


def define_config():
    # Initialize config
    config = config_module.Config

    # Define config
    config.input_path = "input/exafel_1.npz"
    config.output_path = "output/"
    
    config = config_module.Config
    config.compression_ratio = 1000
    config.epochs = 400
    config.early_stopping = False
    config.early_stopping_patience = 100
    config.min_delta = 0
    config.lr_scheduler = True
    config.lr_scheduler_patience = 50
    config.model_name = "CFD_dense_AE"
    config.model_type = "dense"
    config.custom_norm = True
    config.l1 = True
    config.reg_param = 0.001
    config.RHO = 0.05
    config.lr = 0.001
    config.batch_size = 75
    config.test_size = 0.2
    config.data_dimension = 2
    config.apply_normalization = False
    config.deterministic_algorithm = False
    config.compress_to_latent_space = False
    config.convert_to_blocks = [1, 150, 150]
    config.verbose = False


    # FPGA stuff
    config.number_of_columns = 22500  # FIXME: this is doesn't need to be hardcoded
    config.latent_space_size = 225  # FIXME: this is doesn't need to be hardcoded
    config.default_reuse_factor = 1
    config.default_precision = "ap_fixed<16,8>"
    config.Strategy = "latency"
    config.Part = "xcvu9p-flga2104-2L-e"
    config.ClockPeriod = 5
    config.IOType = "io_parallel"
    config.InputShape = (1, 16)
    config.ProjectName = "tiny_test_model"
    config.OutputDir = "workspaces/FPGA_compression_workspace/first_FPGA_Compression_project/output/hls4ml"
    config.InputData = None
    config.OutputPredictions = None
    config.csim = False
    config.synth = True
    config.cosim = False
    config.export = False

    return config


def train(config):
    # Run training
    model, normalization_features, loss_data = trainer_module.run(
        config.input_path, config
    )
    torch.save(model.state_dict(), config.output_path + "compressed_output/model.pt")
    np.save(
        config.output_path + "compressed_output/normalization_features.npy",
        normalization_features,
    )
    helper.loss_plot(loss_data[0], config.output_path, config)


def compress(config):
    # Run compression
    normalization_features = np.load(
        config.output_path + "compressed_output/normalization_features.npy"
    )
    compressed, names, original_shape = compressor_module.run(
        config.input_path,
        config.output_path + "compressed_output/model.pt",
        normalization_features,
        config,
    )

    # Save compressed file to disk
    np.savez_compressed(
        config.output_path + "compressed_output/compressed.npz",
        data=compressed,
        names=names,
        normalization_features=normalization_features,
        original_shape=original_shape,
    )
    # np.save(config.output_path + "loss_data.npy", loss_data)


def decompress(config):
    # Run decompression
    decompressed, names, original_shape = decompressor_module.run(
        config.output_path + "compressed_output/model.pt",
        config.output_path + "compressed_output/compressed.npz",
        config,
    )

    # Save decompressed file to disk
    np.savez(
        config.output_path + "decompressed_output/decompressed.npz",
        data=decompressed,
        names=names,
    )


def plot(config):
    data = np.load(config.input_path)["data"]
    data_decompressed = np.load(
        config.output_path + "decompressed_output/decompressed.npz"
    )["data"].reshape(data.shape[0], data.shape[1], data.shape[2])

    print("Making GIF: ")
    with imageio.get_writer(config.output_path + "/movie.gif", mode="I") as writer:
        for i in tqdm(range(0, 74)):
            figg = helper.plot2D(data[i], data_decompressed[i])
            image_name = config.output_path + f"plot_output/{i}.png"
            plt.savefig(image_name)
            image = imageio.imread(image_name)
            writer.append_data(image)


def FPGA(config):
    baler_helper.perform_hls4ml_conversion(config.output_path, config)


def main():
    config = define_config()

    if sys.argv[1] == "train":
        train(config)
    elif sys.argv[1] == "compress":
        compress(config)
    elif sys.argv[1] == "decompress":
        decompress(config)
    elif sys.argv[1] == "plot":
        plot(config)
    elif sys.argv[1] == "FPGA":
        FPGA(config)
    elif sys.argv[1] == "all":
        train(config)
        compress(config)
        decompress(config)
        plot(config)
        # FPGA(config)
    else:
        print("Unknown argument")


if __name__ == "__main__":
    main()



import os
import random
import time
import sys

import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm
import math

import baler_compressor.helper as helper
import baler_compressor.utils as utils
import baler_compressor.data_processing as data_processing
import baler_compressor.models as models


def run(data_path, config):
    """Main function calling the training functions, ran when --mode=train is selected.
        The three functions called are: `helper.process`, `helper.mode_init` and `helper.training`.

        Depending on `config.data_dimensions`, the calculated latent space size will differ.

    Args:
        output_path (path): Selects base path for determining output path
        config (dataClass): Base class selecting user inputs
        verbose (bool): If True, prints out more information

    Raises:
        NameError: Baler currently only supports 1D (e.g. HEP) or 2D (e.g. CFD) data as inputs.
    """

    config.input_path = data_path
    verbose = config.verbose

    (
        train_set_norm,
        test_set_norm,
        normalization_features,
        original_shape,
    ) = helper.process(
        config.input_path,
        config.custom_norm,
        config.test_size,
        config.apply_normalization,
        config.convert_to_blocks if hasattr(config, "convert_to_blocks") else None,
        verbose,
    )

    if verbose:
        print("Training and testing sets normalized")

    try:
        n_features = 0
        if config.data_dimension == 1:
            number_of_columns = train_set_norm.shape[1]
            config.latent_space_size = math.ceil(
                number_of_columns / config.compression_ratio
            )
            config.number_of_columns = number_of_columns
            n_features = number_of_columns
        elif config.data_dimension == 2:
            if config.model_type == "dense":
                number_of_rows = train_set_norm.shape[1]
                number_of_columns = train_set_norm.shape[2]
                n_features = number_of_columns * number_of_rows
            else:
                number_of_rows = original_shape[1]
                number_of_columns = original_shape[2]
                n_features = number_of_columns
            config.latent_space_size = math.ceil(
                (number_of_rows * number_of_columns) / config.compression_ratio
            )
            config.number_of_columns = number_of_columns
        else:
            raise NameError(
                "Data dimension can only be 1 or 2. Got config.data_dimension value = "
                + str(config.data_dimension)
            )
    except AttributeError:
        if verbose:
            print(
                f"{config.number_of_columns} -> {config.latent_space_size} dimensions"
            )
        assert number_of_columns == config.number_of_columns

    if verbose:
        print(
            f"Intitalizing Model with Latent Size - {config.latent_space_size} and Features - {n_features}"
        )

    device = helper.get_device()
    if verbose:
        print(f"Device used for training: {device}")

    model_object = helper.model_init(config.model_name)
    model = model_object(n_features=n_features, z_dim=config.latent_space_size)
    model.to(device)

    if config.model_name == "Conv_AE_3D" and hasattr(
        config, "compress_to_latent_space"
    ):
        model.set_compress_to_latent_space(config.compress_to_latent_space)

    if verbose:
        print(f"Model architecture:\n{model}")

    # training_path = os.path.join(output_path, "training")
    # if verbose:
    #    print(f"Training path: {training_path}")

    trained_model, loss_data = train(
        model,
        number_of_columns,
        train_set_norm,
        test_set_norm,
        # training_path,
        config,
    )

    if verbose:
        print("Training complete")

    return trained_model, normalization_features, loss_data

    # if config.apply_normalization:
    #    np.save(
    #        os.path.join(training_path, "normalization_features.npy"),
    #        normalization_features,
    #    )
    # if verbose:
    #    print(
    #        f"Normalization features saved to {os.path.join(training_path, 'normalization_features.npy')}"
    #    )
    # helper.model_saver(
    #    trained_model, os.path.join(output_path, "compressed_output", "model.pt")
    # )
    # if verbose:
    #    print(
    #        f"Model saved to {os.path.join(output_path, 'compressed_output', 'model.pt')}"
    #    )


def fit(
    config,
    model,
    train_dl,
    model_children,
    regular_param,
    optimizer,
    latent_dim,
    RHO,
    l1,
    n_dimensions,
):
    """This function trains the model on the train set. It computes the losses and does the backwards propagation, and updates the optimizer as well.
    Args:
        model (modelObject): The model you wish to train
        train_dl (torch.DataLoader): Defines the batched data which the model is trained on
        model_children (list): List of model parameters
        regular_param (float): Determines proportionality constant for the gradient descent step.
        optimizer (torch.optim): Chooses optimizer for gradient descent.
        RHO (float): Float used for KL Divergence (Not currently a feature)
        l1 (boolean): If `True`, use L1 regularization. Otherwise, don't.
        n_dimensions (int): Number of dimensions.
    Returns:
        list, model object: Training loss and trained model
    """

    print("### Beginning Training")

    model.train()

    running_loss = 0.0
    device = helper.get_device()

    for idx, inputs in enumerate(tqdm(train_dl)):
        inputs = inputs.to(device)

        # Set the gradients to zero
        optimizer.zero_grad()

        # Compute the predicted outputs from the input data
        reconstructions = model(inputs)

        if (
            hasattr(config, "custom_loss_function")
            and config.custom_loss_function == "loss_function_swae"
        ):
            z = model.encode(inputs)
            loss, mse_loss, l1_loss = utils.loss_function_swae(
                inputs, z, reconstructions, latent_dim
            )
        else:
            # Compute how far off the prediction is
            loss, mse_loss, l1_loss = utils.mse_sum_loss_l1(
                model_children=model_children,
                true_data=inputs,
                reconstructed_data=reconstructions,
                reg_param=regular_param,
                validate=True,
            )

        # Compute the loss-gradient with
        loss.backward()

        # Update the optimizer
        optimizer.step()

        running_loss += loss.item()

    epoch_loss = running_loss / (idx + 1)
    print(f"# Finished. Training Loss: {loss:.6f}")
    return epoch_loss, mse_loss, l1_loss, model


def validate(model, test_dl, model_children, reg_param):
    """Function used to validate the training. Not necessary for doing compression, but gives a good indication of wether the model selected is a good fit or not.
    Args:
        model (modelObject): Defines the model one wants to validate. The model used here is passed directly from `fit()`.
        test_dl (torch.DataLoader): Defines the batched data which the model is validated on
        model_children (list): List of model parameters
        regular_param (float): Determines proportionality constant for the gradient descent step.
    Returns:
        float: Validation loss
    """
    print("### Beginning Validating")

    model.eval()

    running_loss = 0.0
    device = helper.get_device()

    with torch.no_grad():
        for idx, inputs in enumerate(tqdm(test_dl)):
            inputs = inputs.to(device)
            reconstructions = model(inputs)

            loss, _, _ = utils.mse_sum_loss_l1(
                model_children=model_children,
                true_data=inputs,
                reconstructed_data=reconstructions,
                reg_param=reg_param,
                validate=True,
            )
            running_loss += loss.item()

    epoch_loss = running_loss / (idx + 1)
    print(f"# Finished. Validation Loss: {loss:.6f}")
    return epoch_loss


def seed_worker(worker_id):
    """PyTorch implementation to fix the seeds
    Args:
        worker_id ():
    """
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


def train(model, variables, train_data, test_data, config):
    """Does the entire training loop by calling the `fit()` and `validate()`. Appart from this, this is the main function where the data is converted
        to the correct type for it to be trained, via `torch.Tensor()`. Furthermore, the batching is also done here, based on `config.batch_size`,
        and it is the `torch.utils.data.DataLoader` doing the splitting.
        Applying either `EarlyStopping` or `LR Scheduler` is also done here, all based on their respective `config` arguments.
        For reproducibility, the seeds can also be fixed in this function.
    Args:
        model (modelObject): The model you wish to train
        variables (_type_): _description_
        train_set (ndarray): Array consisting of the train set
        test_set (ndarray): Array consisting of the test set
        project_path (string): Path to the project directory
        config (dataClass): Base class selecting user inputs
    Returns:
        modelObject: fully trained model ready to perform compression and decompression
    """
    # Fix the random seed - TODO: add flag to make this optional

    if config.deterministic_algorithm:
        random.seed(0)
        torch.manual_seed(0)
        np.random.seed(0)
        torch.use_deterministic_algorithms(True)
        g = torch.Generator()
        g.manual_seed(0)

    test_size = config.test_size
    learning_rate = config.lr
    bs = config.batch_size
    reg_param = config.reg_param
    rho = config.RHO
    l1 = config.l1
    epochs = config.epochs
    latent_space_size = config.latent_space_size
    # intermittent_model_saving = config.intermittent_model_saving
    # intermittent_saving_patience = config.intermittent_saving_patience

    model_children = list(model.children())

    # Initialize model with appropriate device
    device = helper.get_device()
    model = model.to(device)

    # Converting data to tensors
    if config.data_dimension == 2:
        if config.model_type == "dense":
            # print(train_data.shape)
            # print(test_data.shape)
            # sys.exit()
            train_ds = torch.tensor(
                train_data, dtype=torch.float32, device=device
            ).view(train_data.shape[0], train_data.shape[1] * train_data.shape[2])
            valid_ds = torch.tensor(test_data, dtype=torch.float32, device=device).view(
                test_data.shape[0], test_data.shape[1] * test_data.shape[2]
            )
        elif config.model_type == "convolutional" and config.model_name == "Conv_AE_3D":
            train_ds = torch.tensor(
                train_data, dtype=torch.float32, device=device
            ).view(
                train_data.shape[0] // bs,
                1,
                bs,
                train_data.shape[1],
                train_data.shape[2],
            )
            valid_ds = torch.tensor(test_data, dtype=torch.float32, device=device).view(
                train_data.shape[0] // bs,
                1,
                bs,
                train_data.shape[1],
                train_data.shape[2],
            )
        elif config.model_type == "convolutional":
            train_ds = torch.tensor(
                train_data, dtype=torch.float32, device=device
            ).view(train_data.shape[0], 1, train_data.shape[1], train_data.shape[2])
            valid_ds = torch.tensor(test_data, dtype=torch.float32, device=device).view(
                train_data.shape[0], 1, train_data.shape[1], train_data.shape[2]
            )
    elif config.data_dimension == 1:
        train_ds = torch.tensor(train_data, dtype=torch.float64, device=device)
        valid_ds = torch.tensor(test_data, dtype=torch.float64, device=device)

    # Pushing input data into the torch-DataLoader object and combines into one DataLoader object (a basic wrapper
    # around several DataLoader objects).

    if config.deterministic_algorithm:
        train_dl = DataLoader(
            train_ds,
            batch_size=bs,
            shuffle=False,
            worker_init_fn=seed_worker,
            generator=g,
            drop_last=False,
        )
        valid_dl = DataLoader(
            valid_ds,
            batch_size=bs,
            worker_init_fn=seed_worker,
            generator=g,
            drop_last=False,
        )
    else:
        train_dl = DataLoader(
            train_ds,
            batch_size=bs,
            shuffle=False,
            drop_last=False,
        )
        valid_dl = DataLoader(
            valid_ds,
            batch_size=bs,
            drop_last=False,
        )

    # Select Optimizer
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

    # Activate early stopping
    if config.early_stopping:
        early_stopping = utils.EarlyStopping(
            patience=config.early_stopping_patience, min_delta=config.min_delta
        )  # Changes to patience & min_delta can be made in configs

    # Activate LR Scheduler
    if config.lr_scheduler:
        lr_scheduler = utils.LRScheduler(
            optimizer=optimizer, patience=config.lr_scheduler_patience
        )

    # Training and Validation of the model
    train_loss = []
    val_loss = []
    start = time.time()

    # Registering hooks for activation extraction
    # if config.activation_extraction:
    #    hooks = model.store_hooks()

    for epoch in range(epochs):
        print(f"Epoch {epoch + 1} of {epochs}")

        (
            train_epoch_loss,
            mse_loss_fit,
            regularizer_loss_fit,
            trained_model,
        ) = fit(
            config=config,
            model=model,
            train_dl=train_dl,
            model_children=model_children,
            regular_param=reg_param,
            optimizer=optimizer,
            latent_dim=latent_space_size,
            RHO=rho,
            l1=l1,
            n_dimensions=config.data_dimension,
        )
        train_loss.append(train_epoch_loss)

        if test_size:
            val_epoch_loss = validate(
                model=trained_model,
                test_dl=valid_dl,
                model_children=model_children,
                reg_param=reg_param,
            )
            val_loss.append(val_epoch_loss)
        else:
            val_epoch_loss = train_epoch_loss
            val_loss.append(val_epoch_loss)

        if config.lr_scheduler:
            lr_scheduler(val_epoch_loss)
        if config.early_stopping:
            early_stopping(val_epoch_loss)
            if early_stopping.early_stop:
                break

        ### Implementation to save models & values after every N epochs, where N is stored in 'intermittent_saving_patience':
        # if intermittent_model_saving:
        #    if epoch % intermittent_saving_patience == 0:
        #        path = os.path.join(project_path, f"model_{epoch}.pt")
        #        helper.model_saver(model, path)

    end = time.time()

    # Saving activations values
    # if config.activation_extraction:
    #    activations = diagnostics.dict_to_square_matrix(model.get_activations())
    #    model.detach_hooks(hooks)
    #    np.save(os.path.join(project_path, "activations.npy"), activations)

    print(f"Training took: {(end - start) / 60:.3} minutes")
    # np.save(
    #    os.path.join(config.output_path, "loss_data.npy"),
    #    np.array([train_loss, val_loss]),
    # )
    loss_data = (np.array([train_loss, val_loss]),)

    if config.model_type == "convolutional":
        final_layer = model.get_final_layer_dims()
        np.save(
            os.path.join("output/compressed_output/", "final_layer.npy"),
            np.array(final_layer),
        )

    return trained_model, loss_data



# Copyright 2022 Baler Contributors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch import nn
from torch.nn import functional as F


import torch.utils.data
from torch.nn import functional as F
from torch.autograd import Function
from baler_compressor import helper


class LowerBound(Function):
    @staticmethod
    def forward(ctx, inputs, bound):
        b = torch.ones(inputs.size(), device=inputs.device) * bound
        b = b.to(inputs.device)
        ctx.save_for_backward(inputs, b)
        return torch.max(inputs, b)

    @staticmethod
    def backward(ctx, grad_output):
        inputs, b = ctx.saved_tensors

        pass_through_1 = inputs >= b
        pass_through_2 = grad_output < 0

        pass_through = pass_through_1 | pass_through_2
        return pass_through.type(grad_output.dtype) * grad_output, None


class GDN(nn.Module):
    """Generalized divisive normalization layer.
    y[i] = x[i] / sqrt(beta[i] + sum_j(gamma[j, i] * x[j]^2))
    """

    def __init__(
        self,
        ch,
        inverse=False,
        beta_min=1e-6,
        gamma_init=0.1,
        reparam_offset=2**-18,
    ):
        super(GDN, self).__init__()
        self.inverse = inverse
        self.beta_min = beta_min
        self.gamma_init = gamma_init
        self.device = helper.get_device()
        self.reparam_offset = torch.tensor([reparam_offset], device=self.device)

        self.build(ch)

    def build(self, ch):
        self.pedestal = self.reparam_offset**2
        self.beta_bound = (self.beta_min + self.reparam_offset**2) ** 0.5
        self.gamma_bound = self.reparam_offset

        # Create beta param
        beta = torch.sqrt(torch.ones(ch, device=self.device) + self.pedestal)
        self.beta = nn.Parameter(beta)

        # Create gamma param
        eye = torch.eye(ch, device=self.device)
        g = self.gamma_init * eye
        g = g + self.pedestal
        gamma = torch.sqrt(g)
        self.gamma = nn.Parameter(gamma)

    def forward(self, inputs):
        unfold = False
        if inputs.dim() == 5:
            unfold = True
            bs, ch, d, w, h = inputs.size()
            inputs = inputs.view(bs, ch, d * w, h)

        _, ch, _, _ = inputs.size()

        # Beta bound and reparam
        beta = LowerBound.apply(self.beta, self.beta_bound)
        beta = beta**2 - self.pedestal

        # Gamma bound and reparam
        gamma = LowerBound.apply(self.gamma, self.gamma_bound)
        gamma = gamma**2 - self.pedestal
        gamma = gamma.view(ch, ch, 1, 1)

        # Norm pool calc
        norm_ = nn.functional.conv2d(inputs**2, gamma, beta)
        norm_ = torch.sqrt(norm_)

        # Apply norm
        if self.inverse:
            outputs = inputs * norm_
        else:
            outputs = inputs / norm_

        if unfold:
            outputs = outputs.view(bs, ch, d, w, h)
        return outputs


class AE(nn.Module):
    # This class is a modified version of the original class by George Dialektakis found at
    # https://github.com/Autoencoders-compression-anomaly/Deep-Autoencoders-Data-Compression-GSoC-2021
    # Released under the Apache License 2.0 found at https://www.apache.org/licenses/LICENSE-2.0.txt
    # Copyright 2021 George Dialektakis

    def __init__(self, n_features, z_dim, *args, **kwargs):
        super(AE, self).__init__(*args, **kwargs)

        self.activations = {}

        # encoder
        self.en1 = nn.Linear(n_features, 200, dtype=torch.float64)
        self.en2 = nn.Linear(200, 100, dtype=torch.float64)
        self.en3 = nn.Linear(100, 50, dtype=torch.float64)
        self.en4 = nn.Linear(50, z_dim, dtype=torch.float64)
        # decoder
        self.de1 = nn.Linear(z_dim, 50, dtype=torch.float64)
        self.de2 = nn.Linear(50, 100, dtype=torch.float64)
        self.de3 = nn.Linear(100, 200, dtype=torch.float64)
        self.de4 = nn.Linear(200, n_features, dtype=torch.float64)

        self.n_features = n_features
        self.z_dim = z_dim

    def encode(self, x):
        h1 = F.leaky_relu(self.en1(x))
        h2 = F.leaky_relu(self.en2(h1))
        h3 = F.leaky_relu(self.en3(h2))
        return self.en4(h3)

    def decode(self, z):
        h4 = F.leaky_relu(self.de1(z))
        h5 = F.leaky_relu(self.de2(h4))
        h6 = F.leaky_relu(self.de3(h5))
        out = self.de4(h6)
        return out

    def forward(self, x):
        z = self.encode(x)
        return self.decode(z)

    # Implementation of activation extraction using the forward_hook method

    def get_hook(self, layer_name):
        def hook(model, input, output):
            self.activations[layer_name] = output.detach()

        return hook

    def get_layers(self) -> list:
        return [self.en1, self.en2, self.en3, self.de1, self.de2, self.de3]

    def store_hooks(self) -> list:
        layers = self.get_layers()
        hooks = []
        for i in range(len(layers)):
            hooks.append(layers[i].register_forward_hook(self.get_hook(str(i))))
        return hooks

    def get_activations(self) -> dict:
        for kk in self.activations:
            self.activations[kk] = F.leaky_relu(self.activations[kk])
        return self.activations

    def detach_hooks(self, hooks: list) -> None:
        for hook in hooks:
            hook.remove()


class CFD_dense_AE(nn.Module):
    # This class is a modified version of the original class by George Dialektakis found at
    # https://github.com/Autoencoders-compression-anomaly/Deep-Autoencoders-Data-Compression-GSoC-2021
    # Released under the Apache License 2.0 found at https://www.apache.org/licenses/LICENSE-2.0.txt
    # Copyright 2021 George Dialektakis

    def __init__(self, n_features, z_dim, *args, **kwargs):
        super(CFD_dense_AE, self).__init__(*args, **kwargs)

        self.activations = {}

        # encoder
        self.en1 = nn.Linear(n_features, 200, dtype=torch.float)
        self.en2 = nn.Linear(200, 100, dtype=torch.float)
        self.en3 = nn.Linear(100, 50, dtype=torch.float)
        self.en4 = nn.Linear(50, z_dim, dtype=torch.float)
        # decoder
        self.de1 = nn.Linear(z_dim, 50, dtype=torch.float)
        self.de2 = nn.Linear(50, 100, dtype=torch.float)
        self.de3 = nn.Linear(100, 200, dtype=torch.float)
        self.de4 = nn.Linear(200, n_features, dtype=torch.float)

        self.n_features = n_features
        self.z_dim = z_dim

    def encode(self, x):
        h1 = F.leaky_relu(self.en1(x))
        h2 = F.leaky_relu(self.en2(h1))
        h3 = F.leaky_relu(self.en3(h2))
        return self.en4(h3)

    def decode(self, z):
        h4 = F.leaky_relu(self.de1(z))
        h5 = F.leaky_relu(self.de2(h4))
        h6 = F.leaky_relu(self.de3(h5))
        out = self.de4(h6)
        return out

    def forward(self, x):
        z = self.encode(x)
        return self.decode(z)

    # Implementation of activation extraction using the forward_hook method

    def get_hook(self, layer_name):
        def hook(model, input, output):
            self.activations[layer_name] = output.detach()

        return hook

    def get_layers(self) -> list:
        return [self.en1, self.en2, self.en3, self.de1, self.de2, self.de3]

    def store_hooks(self) -> list:
        layers = self.get_layers()
        hooks = []
        for i in range(len(layers)):
            hooks.append(layers[i].register_forward_hook(self.get_hook(str(i))))
        return hooks

    def get_activations(self) -> dict:
        for kk in self.activations:
            self.activations[kk] = F.leaky_relu(self.activations[kk])
        return self.activations

    def detach_hooks(self, hooks: list) -> None:
        for hook in hooks:
            hook.remove()


class AE_Dropout_BN(nn.Module):
    def __init__(self, n_features, z_dim, *args, **kwargs):
        super(AE_Dropout_BN, self).__init__(*args, **kwargs)

        # encoder
        self.enc_nn = nn.Sequential(
            nn.Linear(n_features, 200, dtype=torch.float64),
            nn.Dropout(p=0.5),
            nn.LeakyReLU(),
            # nn.BatchNorm1d(200,dtype=torch.float64),
            nn.Linear(200, 100, dtype=torch.float64),
            nn.Dropout(p=0.4),
            nn.LeakyReLU(),
            # nn.BatchNorm1d(100,dtype=torch.float64),
            nn.Linear(100, 50, dtype=torch.float64),
            nn.Dropout(p=0.3),
            nn.LeakyReLU(),
            # nn.BatchNorm1d(50,dtype=torch.float64),
            nn.Linear(50, z_dim, dtype=torch.float64),
            nn.Dropout(p=0.2),
            nn.LeakyReLU(),
            # nn.BatchNorm1d(z_dim,dtype=torch.float64)
        )

        # decoder
        self.dec_nn = nn.Sequential(
            nn.Linear(z_dim, 50, dtype=torch.float64),
            # nn.Dropout(p=0.2),
            nn.LeakyReLU(),
            nn.BatchNorm1d(50, dtype=torch.float64),
            nn.Linear(50, 100, dtype=torch.float64),
            # nn.Dropout(p=0.3),
            nn.LeakyReLU(),
            nn.BatchNorm1d(100, dtype=torch.float64),
            nn.Linear(100, 200, dtype=torch.float64),
            # nn.Dropout(p=0.4),
            nn.LeakyReLU(),
            nn.BatchNorm1d(200, dtype=torch.float64),
            nn.Linear(200, n_features, dtype=torch.float64),
            # nn.Dropout(p=0.5),
            nn.BatchNorm1d(n_features, dtype=torch.float64),
            nn.ReLU(),
        )

        self.n_features = n_features
        self.z_dim = z_dim

    def encode(self, x):
        out = self.enc_nn(x)
        return out

    def decode(self, z):
        out = self.dec_nn(z)
        return out

    def forward(self, x):
        z = self.encode(x)
        return self.decode(z)


class Conv_AE(nn.Module):
    def __init__(self, n_features, z_dim, *args, **kwargs):
        super(Conv_AE, self).__init__(*args, **kwargs)

        self.q_z_mid_dim = 2000
        self.q_z_output_dim = 72128
        self.conv_op_shape = None

        # Encoder

        # Conv Layers
        self.q_z_conv = nn.Sequential(
            nn.Conv2d(1, 8, kernel_size=(2, 5), stride=(1), padding=(1)),
            # nn.BatchNorm2d(8),
            nn.ReLU(),
            nn.Conv2d(8, 16, kernel_size=(3), stride=(1), padding=(1)),
            nn.BatchNorm2d(16),
            nn.ReLU(),
            nn.Conv2d(16, 32, kernel_size=(3), stride=(1), padding=(0)),
            # nn.BatchNorm2d(32),
            nn.ReLU(),
        )
        # Flatten
        self.flatten = nn.Flatten(start_dim=1)

        # Linear layers
        self.q_z_lin = nn.Sequential(
            nn.Linear(self.q_z_output_dim, self.q_z_mid_dim),
            nn.ReLU(),
            # nn.BatchNorm1d(self.q_z_output_dim),
            nn.Linear(self.q_z_mid_dim, z_dim),
            nn.ReLU(),
        )

        # Decoder

        # Linear layers
        self.p_x_lin = nn.Sequential(
            nn.Linear(z_dim, self.q_z_mid_dim),
            nn.ReLU(),
            # nn.BatchNorm1d(self.q_z_output_dim),
            nn.Linear(self.q_z_mid_dim, self.q_z_output_dim),
            nn.ReLU()
            # nn.BatchNorm1d(42720)
        )
        # Conv Layers
        self.p_x_conv = nn.Sequential(
            nn.ConvTranspose2d(32, 16, kernel_size=(3), stride=(1), padding=(0)),
            nn.BatchNorm2d(16),
            nn.ReLU(),
            nn.ConvTranspose2d(16, 8, kernel_size=(3), stride=(1), padding=(1)),
            nn.BatchNorm2d(8),
            nn.ReLU(),
            nn.ConvTranspose2d(8, 1, kernel_size=(2, 5), stride=(1), padding=(1)),
        )

    def encode(self, x):
        # Conv
        out = self.q_z_conv(x)
        self.conv_op_shape = out.shape
        self.q_z_output_dim = out.shape[1] * out.shape[2] * out.shape[3]

        # Flatten
        out = self.flatten(out)
        # Dense
        out = self.q_z_lin(out)
        return out

    def decode(self, z):
        # Dense
        out = self.p_x_lin(z)
        # Unflatten
        out = out.view(
            self.conv_op_shape[0],
            self.conv_op_shape[1],
            self.conv_op_shape[2],
            self.conv_op_shape[3],
        )
        # Conv transpose
        out = self.p_x_conv(out)
        return out

    def forward(self, x):
        z = self.encode(x)
        out = self.decode(z)
        return out

    def get_final_layer_dims(self):
        return self.conv_op_shape

    def set_final_layer_dims(self, conv_op_shape):
        self.conv_op_shape = conv_op_shape


class FPGA_prototype_model(nn.Module):
    def __init__(self, n_features, z_dim, *args, **kwargs):
        super(FPGA_prototype_model, self).__init__(*args, **kwargs)

        # encoder
        self.en1 = nn.Linear(n_features, 20, dtype=torch.float64)
        self.en_act1 = nn.ReLU()
        self.en2 = nn.Linear(20, 10, dtype=torch.float64)
        self.en_act2 = nn.ReLU()
        self.en3 = nn.Linear(10, z_dim, dtype=torch.float64)

        # decoder
        self.de1 = nn.Linear(z_dim, 10, dtype=torch.float64)
        self.de_act1 = nn.ReLU()
        self.de2 = nn.Linear(10, 20, dtype=torch.float64)
        self.de_act2 = nn.ReLU()
        self.de3 = nn.Linear(20, n_features, dtype=torch.float64)

        self.n_features = n_features
        self.z_dim = z_dim

    def encode(self, x):
        s1 = self.en1(x)
        s2 = self.en_act1(s1)
        s3 = self.en2(s2)
        s4 = self.en_act2(s3)
        s5 = self.en3(s4)
        return s5

    def decode(self, z):
        s6 = self.de1(z)
        s7 = self.de_act1(s6)
        s8 = self.de2(s7)
        s9 = self.de_act2(s8)
        s10 = self.de3(s9)
        return s10

    def forward(self, x):
        z = self.encode(x)
        return self.decode(z)

    def get_layers(self) -> list:
        return [
            self.en1,
            self.en_act1,
            self.en2,
            self.en_act2,
            self.en3,
            self.de1,
            self.de_act1,
            self.de2,
            self.de_act2,
            self.de3,
        ]


class Conv_AE_3D(nn.Module):
    def __init__(self, n_features, z_dim, *args, **kwargs):
        super(Conv_AE_3D, self).__init__(*args, **kwargs)

        self.q_z_mid_dim = 2000
        self.q_z_output_dim = 612864
        self.compress_to_latent_space = True
        self.debug = False

        # Encoder

        # Conv Layers
        self.q_z_conv = nn.Sequential(
            nn.Conv3d(1, 4, kernel_size=(1, 3, 5), stride=(1), padding=(0)),
            # nn.BatchNorm2d(8),
            nn.ReLU(),
            nn.Conv3d(4, 8, kernel_size=(3), stride=(1), padding=(0)),
            # nn.BatchNorm2d(16),
            nn.ReLU(),
            nn.Conv3d(8, 16, kernel_size=(3), stride=(1), padding=(0)),
            # nn.BatchNorm2d(16),
            nn.ReLU(),
            nn.Conv3d(16, 32, kernel_size=(3), stride=(1), padding=(0)),
            # nn.BatchNorm2d(16),
            nn.ReLU(),
            nn.Conv3d(32, 64, kernel_size=(3), stride=(1), padding=(0)),
            # nn.BatchNorm2d(16),
            nn.ReLU(),
            nn.Conv3d(64, 64, kernel_size=(1, 3, 3), stride=(1), padding=(0)),
            # nn.BatchNorm2d(16),
            nn.ReLU(),
        )
        # Flatten
        self.flatten = nn.Flatten(start_dim=1)

        # Linear layers
        self.q_z_lin = nn.Sequential(
            nn.Linear(self.q_z_output_dim, self.q_z_mid_dim),
            nn.ReLU(),
            # nn.BatchNorm1d(self.q_z_mid_dim),
            nn.Linear(self.q_z_mid_dim, z_dim),
            nn.ReLU(),
        )

        # Decoder

        # Linear layers
        self.p_x_lin = nn.Sequential(
            nn.Linear(z_dim, self.q_z_mid_dim),
            nn.ReLU(),
            # nn.BatchNorm1d(self.q_z_mid_dim),
            nn.Linear(self.q_z_mid_dim, self.q_z_output_dim),
            nn.ReLU(),
        )

        # Conv Layers
        self.p_x_conv = nn.Sequential(
            nn.ConvTranspose3d(64, 64, kernel_size=(1, 3, 3), stride=(1), padding=(0)),
            nn.BatchNorm3d(64),
            nn.ReLU(),
            nn.ConvTranspose3d(64, 32, kernel_size=(3), stride=(1), padding=(0)),
            nn.BatchNorm3d(32),
            nn.ReLU(),
            nn.ConvTranspose3d(32, 16, kernel_size=(3), stride=(1), padding=(0)),
            nn.BatchNorm3d(16),
            nn.ReLU(),
            nn.ConvTranspose3d(16, 8, kernel_size=(3), stride=(1), padding=(0)),
            nn.BatchNorm3d(8),
            nn.ReLU(),
            nn.ConvTranspose3d(8, 4, kernel_size=(3), stride=(1), padding=(0)),
            nn.BatchNorm3d(4),
            nn.ReLU(),
            nn.ConvTranspose3d(4, 1, kernel_size=(1, 3, 5), stride=(1), padding=(0)),
        )

    def encode(self, x):
        # Conv

        out = self.q_z_conv(x)
        # Flatten
        out = self.flatten(out)

        if self.compress_to_latent_space:
            # Dense
            out = self.q_z_lin(out)

        return out

    def decode(self, out):
        if self.compress_to_latent_space:
            out = self.p_x_lin(out)

        out = out.view(4, 64, 7, 38, 36)

        # Conv transpose
        out = self.p_x_conv(out)

        return out

    def forward(self, x):
        z = self.encode(x)
        out = self.decode(z)
        return out

    def set_compress_to_latent_space(self, compress_to_latent_space):
        self.compress_to_latent_space = compress_to_latent_space


class Conv_AE_GDN(nn.Module):
    def __init__(self, n_features, z_dim, *args, **kwargs):
        super(Conv_AE_GDN, self).__init__(*args, **kwargs)

        self.q_z_mid_dim = 2000
        self.q_z_output_dim = 128
        self.conv_op_shape = None

        # Encoder

        # Conv Layers
        self.q_z_conv = nn.Sequential(
            nn.Conv2d(1, 8, kernel_size=(2, 5), stride=(1), padding=(1)),
            # nn.BatchNorm2d(8),
            GDN(8),
            nn.Conv2d(8, 16, kernel_size=(3), stride=(1), padding=(1)),
            nn.BatchNorm2d(16),
            GDN(16),
            nn.Conv2d(16, 32, kernel_size=(3), stride=(1), padding=(0)),
            # nn.BatchNorm2d(32),
            GDN(32),
        )
        # Flatten
        self.flatten = nn.Flatten(start_dim=1)

        # Linear layers
        self.q_z_lin = nn.Sequential(
            nn.Linear(self.q_z_output_dim, self.q_z_mid_dim),
            nn.ReLU(),
            # nn.BatchNorm1d(self.q_z_output_dim),
            nn.Linear(self.q_z_mid_dim, z_dim),
            nn.ReLU(),
        )

        # Decoder

        # Linear layers
        self.p_x_lin = nn.Sequential(
            nn.Linear(z_dim, self.q_z_mid_dim),
            nn.ReLU(),
            # nn.BatchNorm1d(self.q_z_output_dim),
            nn.Linear(self.q_z_mid_dim, self.q_z_output_dim),
            nn.ReLU()
            # nn.BatchNorm1d(42720)
        )
        # Conv Layers
        self.p_x_conv = nn.Sequential(
            nn.ConvTranspose2d(32, 16, kernel_size=(3), stride=(1), padding=(0)),
            # nn.BatchNorm2d(16),
            GDN(16, inverse=True),
            nn.ConvTranspose2d(16, 8, kernel_size=(3), stride=(1), padding=(1)),
            # nn.BatchNorm2d(8),
            GDN(8, inverse=True),
            nn.ConvTranspose2d(8, 1, kernel_size=(2, 5), stride=(1), padding=(1)),
        )

    def encode(self, x):
        # Conv
        out = self.q_z_conv(x)
        self.conv_op_shape = out.shape
        self.q_z_output_dim = out.shape[1] * out.shape[2] * out.shape[3]

        # Flatten
        out = self.flatten(out)
        # Dense
        out = self.q_z_lin(out)
        return out

    def decode(self, z):
        # Dense
        out = self.p_x_lin(z)
        # Unflatten
        out = out.view(
            self.conv_op_shape[0],
            self.conv_op_shape[1],
            self.conv_op_shape[2],
            self.conv_op_shape[3],
        )
        # Conv transpose
        out = self.p_x_conv(out)
        return out

    def forward(self, x):
        z = self.encode(x)
        out = self.decode(z)
        return out

    def get_final_layer_dims(self):
        return self.conv_op_shape

    def set_final_layer_dims(self, conv_op_shape):
        self.conv_op_shape = conv_op_shape



import os
import time
from math import ceil

import numpy as np

import baler_compressor.helper as helper
import gzip
import sys


def run(model_path, compressed_output_path, config):
    """Main function calling the decompression functions, ran when --mode=decompress is selected.
    The main function being called here is: `helper.decompress`

        If `config.apply_normalization=True` the output is un-normalized with the same normalization features saved from `perform_training()`.

    Args:
        output_path (path): Selects base path for determining output path
        config (dataClass): Base class selecting user inputs
        verbose (bool): If True, prints out more information
    """
    print("Decompressing...")

    output_path = compressed_output_path
    verbose = config.verbose

    start = time.time()
    model_name = config.model_name
    # data_before = np.load(config.input_path)["data"]
    decompressed, names, normalization_features, original_shape = helper.decompress(
        model_path=model_path,
        input_path=compressed_output_path,
        input_path_deltas=os.path.join(
            output_path, "compressed_output", "compressed_deltas.npz.gz"
        ),
        input_batch_index=os.path.join(
            output_path,
            "compressed_output",
            "compressed_batch_index_metadata.npz.gz",
        ),
        model_name=model_name,
        config=config,
        output_path=output_path,
    )
    if verbose:
        print(f"Model used: {model_name}")

    # if hasattr(config, "convert_to_blocks") and config.convert_to_blocks:
    #     print(
    #         "Converting Blocked Data into Standard Format. Old Shape - ",
    #         decompressed.shape,
    #         "Target Shape - ",
    #         data_before.shape,
    #     )
    #     if config.model_type == "dense":
    #         decompressed = decompressed.reshape(
    #             data_before.shape[0], data_before.shape[1], data_before.shape[2]
    #         )
    #     else:
    #         decompressed = decompressed.reshape(
    #             data_before.shape[0], 1, data_before.shape[1], data_before.shape[2]
    #         )

    if config.apply_normalization:
        print("Un-normalizing...")
        normalization_features = np.load(
            os.path.join(
                "/".join(output_path.split("/")[:-1]),
                "normalization_features.npy",
            ),
        )
        if verbose:
            print(
                f"Normalization features loaded from {os.path.join(output_path, 'training', 'normalization_features.npy')}"
            )

        decompressed = helper.renormalize(
            decompressed,
            normalization_features[0],
            normalization_features[1],
        )

    try:
        if verbose:
            print("Converting to original data types")
        type_list = config.type_list
        decompressed = np.transpose(decompressed)
        for index, column in enumerate(decompressed):
            decompressed[index] = decompressed[index].astype(type_list[index])
        decompressed = np.transpose(decompressed)
    except AttributeError:
        pass

    end = time.time()
    print("Decompression took:", f"{(end - start) / 60:.3} minutes")

    return decompressed, names, original_shape

    # if config.extra_compression:
    #     if verbose:
    #         print("Extra compression selected")
    #         print(
    #             f"Saving decompressed file to {os.path.join(output_path, 'decompressed_output', 'decompressed.npz')}"
    #         )
    #     np.savez_compressed(
    #         os.path.join(output_path, "decompressed_output", "decompressed.npz"),
    #         data=decompressed,
    #         names=names,
    #     )
    # else:
    #     np.savez(
    #         os.path.join(output_path, "decompressed_output", "decompressed.npz"),
    #         data=decompressed,
    #         names=names,
    #     )



import numpy as np
import matplotlib.pyplot as plt
import os


def plot2D(original, reconstructed):
    plt.close()
    diff = original - reconstructed

    max_value = np.amax([np.amax(original), np.amax(reconstructed)])
    min_value = np.amin([np.amin(original), np.amin(reconstructed)])

    fig, axs = plt.subplots(
        1, 3, figsize=(29.7 * (1 / 2.54), 10 * (1 / 2.54)), sharey=False
    )
    axs[0].set_title("Original", fontsize=11)
    im1 = axs[0].imshow(original, vmax=max_value, vmin=min_value)
    axs[0].invert_yaxis()

    axs[1].set_title("Reconstructed", fontsize=11)
    im2 = axs[1].imshow(reconstructed, vmax=max_value, vmin=min_value)
    axs[1].invert_yaxis()

    axs[2].set_title("Difference", fontsize=11)
    im3 = axs[2].imshow(diff, vmax=max_value, vmin=min_value)
    axs[2].invert_yaxis()

    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([0.815, 0.2, 0.02, 0.59])
    cb2 = fig.colorbar(im3, cax=cbar_ax, aspect=10)
    return fig


def loss_plot(loss_data, output_path, config):
    """This function Plots the loss from the training and saves it

    Args:
        path_to_loss_data (string): Path to file containing loss plot data generated during training
        output_path (path): Directory path to which the loss plot is saved
        config (dataclass): The config class containing attributes set in the config file
    """
    # loss_data = np.load(path_to_loss_data)
    str_list = ["Epochs:", "Model Name:", "Reg. Param:", "lr:", "BS:"]

    train_loss = loss_data[0]
    val_loss = loss_data[1]
    conf_list = [
        len(train_loss),
        config.model_name,
        config.reg_param,
        config.lr,
        config.batch_size,
    ]

    plt.figure(figsize=(10, 7))
    plt.title("Loss plot")
    plt.plot(train_loss, color="orange", label="Train Loss")
    if config.test_size:
        plt.plot(val_loss, color="red", label="Validation Loss")
    for i in range(len(conf_list)):
        plt.plot([], [], " ", label=str_list[i] + " " + str(conf_list[i]))
    plt.xlabel("Epochs")
    plt.yscale("log")
    plt.ylabel("Loss")
    plt.legend(loc="best")
    plt.savefig(os.path.join(output_path, "Loss_plot.png"))
    # plt.show()



# Copyright 2022 Baler Contributors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple

import numpy as np
import torch
from numpy import ndarray
from sklearn.model_selection import train_test_split

from baler_compressor import helper
from baler_compressor import models
from numba import jit


@jit(cache=True, parallel=True, nopython=True)
def convert_to_blocks_util(blocks, data:np.float32):
    print(
        "Converted Dataset to Blocks of Size - ", blocks, " from original ", data.shape
    )
    blocks = np.array(blocks)
    original_shape = np.array(data.shape)
    total_size = np.prod(original_shape)
    data = data.reshape((total_size // (blocks[1] * blocks[2])), blocks[1], blocks[2])
    return data


def save_model(model, model_path: str) -> None:
    """Saves the models state dictionary as a `.pt` file to the given path.

    Args:
        model (nn.Module): The PyTorch model to save.
        model_path (str): String defining the models save path.

    Returns:
        None: Saved model state dictionary as `.pt` file.
    """
    torch.save(model.state_dict(), model_path)


def initialise_model(model_name: str):
    """Initializing the models attributes to a model_object variable.

    Args:
        model_name (str): The name of the model you wish to initialize. This should correspond to what your Model name.

    Returns:
        class: Object with the models class attributes
    """
    model_object = getattr(models, model_name)
    return model_object


def load_model(model_object, model_path: str, n_features: int, z_dim: int):
    """Loads the state dictionary of the trained model into a model variable. This variable is then used for passing
    data through the encoding and decoding functions.

    Args:
        model_object (object): Object with the models attributes
        model_path (str): Path to model
        n_features (int): Input dimension size
        z_dim (int): Latent space size

    Returns: nn.Module: Returns a model object with the attributes of the model class, with the selected state
    dictionary loaded into it.
    """
    device = helper.get_device()
    model = model_object(n_features, z_dim)
    model.to(device)

    # Loading the state_dict into the model
    model.load_state_dict(
        torch.load(str(model_path), map_location=device), strict=False
    )
    return model


@jit(cache=True, parallel=True, forceobj=True)
def find_minmax(data):
    """Obtains the minimum and maximum values for each column.

    Args:
        data (ndarray): Any dataset as a `ndarray` which one eventually wants to normalize using the Min-Max method.

    Returns: ndarray: An array of two lists. One of the lists contains the minimum of each column, while the other
    list contains `feature_range = max - min` for each column.
    """
    data = list(data)
    true_max_list = np.apply_along_axis(np.max, axis=0, arr=data)
    true_min_list = np.apply_along_axis(np.min, axis=0, arr=data)

    # Computes the range
    feature_range_list = true_max_list - true_min_list

    normalization_features = np.array([true_min_list, feature_range_list])
    return normalization_features


@jit(cache=True, parallel=True, forceobj=True)
def normalize(data, custom_norm: bool):
    """This function scales the data to be in the range [0,1], based on the Min Max normalization method. It finds
    the minimum and maximum values of each column and computes the values according to: x_norm = (x - x_min) / (x_max
    - x_min).

    Args: data (ndarray): A dataset of type `ndarray`. custom_norm (boolean): If you want to do Min Max normalization
    or any custom normalization. Custom normalization is not supported at the moment.

    Returns: ndarray: If not custom_norm: Input data where every column is scaled to be in the range [0,
    1]. Otherwise, the input data is returned
    """
    data = np.array(data)
    if custom_norm:
        pass
    elif not custom_norm:
        true_min = np.min(data)
        true_max = np.max(data)
        feature_range = true_max - true_min
        data = [((i - true_min) / feature_range) for i in data]
        data = np.array(data)
    return data


@jit(cache=True, parallel=True, nopython=True)
def split(data, test_size: float, random_state: int) -> Tuple[ndarray, ndarray]:
    """Splits the given data according to a test size into two sets, train and test. This is a sklearn function,
    opted from: https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html

    Args: data (ndarray): Dataset which you want to split test_size (float): Value between 0.0 and 1.0 determining
    how large the test size will be. Example: test_size = 0.15 randomly selects 15% of the original dataset and puts
    it into a new array. random_state (int): Shuffling control. Passing an int here shuffles the data the same way
    every time. Used for reproducibility.

    Returns:
        ndarray, ndarray: Two datasets split in size according to the `test_size` value.
    """
    return train_test_split(data, test_size=test_size, random_state=random_state)


@jit(cache=True, parallel=True, forceobj=True)
def renormalize_std(
    input_data: ndarray, true_min: float, feature_range: float
) -> ndarray:
    """Computes the un-normalization of normalized arrays. This function is used in combination with
    `renormalize_func` to perform this operation on an entire dataset.

    Args:
        input_data (ndarray): Normalized data you want to un-normalize from the range [0,1]
        true_min (float): A columns minimum
        feature_range (float): A columns feature range, computed as `max - min`

    Returns:
        ndarray: Un-normalized data array
    """
    return np.array([((i * feature_range) + true_min) for i in list(input_data)])


@jit(cache=True, parallel=True, forceobj=True)
def renormalize_func(norm_data: ndarray, min_list: List, range_list: List) -> ndarray:
    """Un-normalizes an entire dataset. Applies the `renormalize_std`function across an entire dataset.
        `min_list` and `range_list` are obtained from the `normalization_features.npy` file.

    Args:
        norm_data (ndarray): Normalized data which one wants to un-normalized.
        min_list (list): List of minimum values. Should have the same length as the amount of columns.
        range_list (list): List of feature range values. Should have the same length as the amount of columns.

    Returns:
        ndarray: Array with the un-normalized values.
    """
    norm_data = np.array(norm_data)
    min_list = np.array(min_list)
    range_list = np.array(range_list)
    return norm_data * range_list + min_list
